use crate::{err, page, station::Station, succ};
use axum::{
    extract::State,
    response::{IntoResponse, Redirect},
    routing::{get, post},
    Form, Router,
};
use maud::{html, Markup, PreEscaped};
use serde::{Deserialize, Serialize};
use sqlx::{FromRow, SqlitePool};
use std::sync::Arc;
use tower_sessions::Session;

#[derive(FromRow, Debug, Serialize, Deserialize)]
pub(crate) struct Route {
    pub(crate) id: i64,
    name: String,
}

impl Route {
    async fn all(db: &SqlitePool) -> Vec<Self> {
        sqlx::query_as::<_, Self>("SELECT id, name FROM route;")
            .fetch_all(db)
            .await
            .unwrap()
    }

    pub async fn find_by_id(db: &SqlitePool, id: i64) -> Option<Self> {
        sqlx::query_as!(Self, "SELECT id, name FROM route WHERE id = ?", id)
            .fetch_one(db)
            .await
            .ok()
    }

    async fn create(db: &SqlitePool, name: &str) -> Result<(), String> {
        sqlx::query!("INSERT INTO route(name) VALUES (?)", name)
            .execute(db)
            .await
            .map_err(|e| e.to_string())?;
        Ok(())
    }

    async fn update_name(&self, db: &SqlitePool, name: &str) {
        sqlx::query!("UPDATE route SET name = ? WHERE id = ?", name, self.id)
            .execute(db)
            .await
            .unwrap();
    }

    async fn add_station(&self, db: &SqlitePool, station: &Station) -> Result<(), String> {
        sqlx::query!(
            r#"
        INSERT INTO route_station (route_id, station_id, pos)
        VALUES (?, ?, (
            SELECT COALESCE(MAX(pos), 0) + 2 
            FROM route_station 
            WHERE route_id = ?
        ))
        "#,
            self.id,
            station.id,
            self.id
        )
        .execute(db)
        .await
        .map_err(|e| e.to_string())?;

        Ok(())
    }

    async fn delete_station(&self, db: &SqlitePool, station: &Station) -> bool {
        let result = sqlx::query!(
            "DELETE FROM route_station WHERE route_id = ? AND station_id = ?",
            self.id,
            station.id
        )
        .execute(db)
        .await
        .unwrap();

        result.rows_affected() > 0
    }

    async fn move_station_higher(&self, db: &SqlitePool, station: &Station) -> bool {
        let result = sqlx::query!(
            "UPDATE route_station SET pos = pos-3 WHERE route_id = ? AND station_id = ?",
            self.id,
            station.id
        )
        .execute(db)
        .await
        .unwrap();

        if result.rows_affected() == 0 {
            return false;
        }

        sqlx::query(
            &format!(
            "
-- Step 1: Create a temporary table with new rank values
CREATE TEMP TABLE IF NOT EXISTS temp_pos AS
SELECT
  route_id,
  station_id,
  ROW_NUMBER() OVER (ORDER BY pos ASC) AS new_rank
FROM
  route_station WHERE route_id = {};

-- Step 2: Update the original table
UPDATE route_station 
SET pos = (SELECT 2*new_rank FROM temp_pos WHERE temp_pos.route_id = route_station.route_id AND temp_pos.station_id = route_station.station_id) WHERE route_id = {};

-- Step 3: Drop the temporary table (no longer needed)
DROP TABLE temp_pos;",
            self.id,
            self.id,
        ))
        .execute(db)
        .await
        .unwrap();

        true
    }

    async fn delete(&self, db: &SqlitePool) -> Result<(), String> {
        sqlx::query!("DELETE FROM route WHERE id = ?", self.id)
            .execute(db)
            .await
            .map_err(|e| e.to_string())?;
        Ok(())
    }

    async fn stations(&self, db: &SqlitePool) -> Vec<Station> {
        sqlx::query_as::<_, Station>(
            r#"
        SELECT s.id, s.name, s.notes, s.amount_people, s.last_login, s.pw, s.lat, s.lng
        FROM station s
        JOIN route_station rs ON s.id = rs.station_id
        WHERE rs.route_id = ?
        ORDER BY rs.pos
        "#,
        )
        .bind(self.id)
        .fetch_all(db)
        .await
        .unwrap()
    }

    async fn stations_not_in_route(&self, db: &SqlitePool) -> Vec<Station> {
        sqlx::query_as::<_, Station>(
            r#"
        SELECT id, name, notes, amount_people, last_login, pw, lat, lng
        FROM station
        WHERE id NOT IN (
            SELECT station_id
            FROM route_station
            WHERE route_id = ?
        )
        ORDER BY name
        "#,
        )
        .bind(self.id)
        .fetch_all(db)
        .await
        .unwrap()
    }
}

async fn index(State(db): State<Arc<SqlitePool>>, session: Session) -> Markup {
    let routes = Route::all(&db).await;

    let content = html! {
        h1 {
            a href="/" { "↩️" }
            "Routen"
        }
        article {
            em { "Routen " }
            "definieren welche "
            a href="/station" { "Stationen" }
            " von den Teilnehmern in welcher Reihenfolge abgeklappert werden sollen. Wenn es verschiedene Kategorien (zB Kinder- und Erwachsenenwertung) gibt, kannst du auch mehrere Routen mit (teils) überlappenden Stationen erstellen."
        }
        ol {
            @for route in &routes{
                li {
                    @if route.stations(&db).await.is_empty() {
                        em data-tooltip="Keine Stationen zugeteilt" {
                            "⚠️"
                        }
                    }
                     a href=(format!("/route/{}", route.id)){
                        (route.name)
                     }
                     a href=(format!("/route/{}/delete", route.id))
                       onclick="return confirm('Bist du sicher, dass die Route gelöscht werden soll? Das kann _NICHT_ mehr rückgängig gemacht werden.');" {
                         "🗑️"
                     }
                }
            }
        }
        @if routes.is_empty() {
            article class="warning" {
                "Es gibt noch keine Routen. Das kannst du hier ändern ⤵️"
            }
        }
        h2 { "Neue Route" }
        form action="/route" method="post" {
            fieldset role="group" {
                input type="text" name="name" placeholder="Routenname" required;
                input type="submit" value="Neue Route";
            }
        }
    };
    page(content, session, false).await
}

#[derive(Deserialize)]
struct CreateForm {
    name: String,
}

async fn create(
    State(db): State<Arc<SqlitePool>>,
    session: Session,
    Form(form): Form<CreateForm>,
) -> impl IntoResponse {
    match Route::create(&db, &form.name).await {
        Ok(_) => succ!(session,"Route '{}' erfolgreich erstellt!", form.name),
        Err(e) => err!(
            session,
            "Route '{}' konnte _NICHT_ erstellt werden, da es bereits eine Route mit diesem Namen gibt ({e})!",
            form.name
        ),
    }

    Redirect::to("/route")
}

async fn delete(
    State(db): State<Arc<SqlitePool>>,
    session: Session,
    axum::extract::Path(id): axum::extract::Path<i64>,
) -> impl IntoResponse {
    let Some(route) = Route::find_by_id(&db, id).await else {
        err!(
            session,
            "Route mit ID {id} konnte nicht gelöscht werden, da sie nicht existiert"
        );

        return Redirect::to("/route");
    };

    match route.delete(&db).await {
        Ok(_) => succ!(session, "Route '{}' erfolgreich gelöscht!", route.name),
        Err(e) => err!(
            session,
            "Route '{}' kann nicht gelöscht werden, da sie bereits verwendet wird. ({e})",
            route.name
        ),
    }

    Redirect::to("/route")
}

async fn view(
    State(db): State<Arc<SqlitePool>>,
    session: Session,
    axum::extract::Path(id): axum::extract::Path<i64>,
) -> Result<Markup, impl IntoResponse> {
    let Some(route) = Route::find_by_id(&db, id).await else {
        err!(
            session,
            "Route mit ID {id} konnte nicht geöffnet werden, da sie nicht existiert"
        );

        return Err(Redirect::to("/route"));
    };

    let cur_stations = route.stations(&db).await;
    let stations_not_in_route = route.stations_not_in_route(&db).await;

    let content = html! {
        h1 {
            a href="/route" { "↩️" }
            "Route " (route.name)
        }
        article {
            details {
                summary { "Routenname bearbeiten ✏️" }
                form action=(format!("/route/{}/name", route.id)) method="post" {
                    input type="text" name="name" value=(route.name) required;
                    input type="submit" value="Speichern";
                }
            }
        }
        h2 { "Stationen" }
        @if cur_stations.is_empty() {
            @if stations_not_in_route.is_empty() {
                article class="error" {
                    (PreEscaped("Bevor du einer Route Stationen zuweisen kannst, musst du die Stationen erstellen &rarr; "))
                    a role="button" href="/station" {
                        "Station erstellen"
                    }
                }
            } @else {
                article class="warning" {
                    "Diese Route hat noch keine Stationen zugewiesen. Das kannst du hier ändern ⤵️"
                }
            }
        } @else {
            ol {
                @for (idx, station) in cur_stations.into_iter().enumerate() {
                    li {
                        (station.name)
                         @if idx > 0 {
                            a href=(format!("/route/{}/move-station-higher/{}", route.id, station.id)){
                                em data-tooltip=(format!("{} nach vor reihen", station.name)) {
                                    "⬆️"
                                }
                            }
                         }
                        a href=(format!("/route/{}/delete-station/{}", route.id, station.id))
                          onclick="return confirm('Bist du sicher, dass die Station von der Route entfernt werden soll?');" {
                            "🗑️"
                        }
                    }
                }
            }
        }
        @if !stations_not_in_route.is_empty(){
            form action=(format!("/route/{}/add-station", route.id)) method="post" {
                select name="station" aria-label="Hinzuzufügende Station auswählen" required {
                    @for station in &stations_not_in_route {
                        option value=(station.id) {
                            (station.name)
                        }
                    }
                    input type="submit" value="Hinzufügen";
                }
        }
        }
    };
    Ok(page(content, session, false).await)
}

#[derive(Deserialize)]
struct UpdateNameForm {
    name: String,
}
async fn update_name(
    State(db): State<Arc<SqlitePool>>,
    session: Session,
    axum::extract::Path(id): axum::extract::Path<i64>,
    Form(form): Form<UpdateNameForm>,
) -> impl IntoResponse {
    let Some(route) = Route::find_by_id(&db, id).await else {
        err!(
            session,
            "Route mit ID {id} konnte nicht geöffnet werden, da sie nicht existiert"
        );

        return Redirect::to("/route");
    };

    route.update_name(&db, &form.name).await;

    succ!(
        session,
        "Route '{}' heißt ab sofort {}.",
        route.name,
        form.name
    );

    Redirect::to(&format!("/route/{id}"))
}

#[derive(Deserialize)]
struct AddStationForm {
    station: i64,
}
async fn add_station(
    State(db): State<Arc<SqlitePool>>,
    session: Session,
    axum::extract::Path(id): axum::extract::Path<i64>,
    Form(form): Form<AddStationForm>,
) -> impl IntoResponse {
    let Some(route) = Route::find_by_id(&db, id).await else {
        err!(
            session,
            "Route mit ID {id} konnte nicht geöffnet werden, da sie nicht existiert"
        );

        return Redirect::to("/route");
    };
    let Some(station) = Station::find_by_id(&db, form.station).await else {
        err!(
            session,
            "Station mit ID {id} konnte nicht hinzugefügt werden, da sie nicht existiert"
        );

        return Redirect::to(&format!("/route/{id}"));
    };

    match route.add_station(&db, &station).await {
        Ok(_) => succ!(
            session,
            "Station {} wurde der Route {} hinzugefügt",
            station.name,
            route.name
        ),
        Err(e) => err!(
            session,
            "Station {} kann nur 1x der Route {} hinzugefügt werden. ({e})",
            station.name,
            route.name
        ),
    }

    Redirect::to(&format!("/route/{id}"))
}

async fn delete_station(
    State(db): State<Arc<SqlitePool>>,
    session: Session,
    axum::extract::Path((route_id, station_id)): axum::extract::Path<(i64, i64)>,
) -> impl IntoResponse {
    let Some(route) = Route::find_by_id(&db, route_id).await else {
        err!(
            session,
            "Konnte keine Station von Route mit ID {route_id} entfernen, da diese Route nicht existiert."
        );

        return Redirect::to("/route");
    };
    let Some(station) = Station::find_by_id(&db, station_id).await else {
        err!(
            session,
            "Konnte Station mit der ID {station_id} nicht von der Route {} entfernen, da die Station nicht existiert.", route.name
        );

        return Redirect::to(&format!("/route/{route_id}"));
    };

    if route.delete_station(&db, &station).await {
        succ!(
            session,
            "Station '{}' wurde von der Route '{}' gelöscht",
            station.name,
            route.name
        );
    } else {
        err!(
            session,
            "Station '{}' konnte nicht von der Route '{}' gelöscht werden, da diese nicht auf dieser Route liegt",
            station.name,
            route.name
        );
    }

    Redirect::to(&format!("/route/{route_id}"))
}

async fn move_station_higher(
    State(db): State<Arc<SqlitePool>>,
    session: Session,
    axum::extract::Path((route_id, station_id)): axum::extract::Path<(i64, i64)>,
) -> impl IntoResponse {
    let Some(route) = Route::find_by_id(&db, route_id).await else {
        err!(
            session,
            "Konnte keine Station von Route mit ID {route_id} verschieben, da diese Route nicht existiert."
        );

        return Redirect::to("/route");
    };
    let Some(station) = Station::find_by_id(&db, station_id).await else {
        err!(
            session,
            "Konnte Station mit der ID {station_id} nicht in der Route {} verschieben, da die Station nicht existiert.", route.name
        );

        return Redirect::to(&format!("/route/{route_id}"));
    };

    if route.move_station_higher(&db, &station).await {
        succ!(
            session,
            "Station '{}' wurde in der Route '{}' erfolgreich vorgereiht",
            station.name,
            route.name
        );
    } else {
        err!(
            session,
            "Station '{}' konnte in der Route '{}' nicht vorgereiht werden, da diese nicht auf dieser Route liegt",
            station.name,
            route.name
        );
    }

    Redirect::to(&format!("/route/{route_id}"))
}

pub(super) fn routes() -> Router<Arc<SqlitePool>> {
    Router::new()
        .route("/", get(index))
        .route("/", post(create))
        .route("/{id}/delete", get(delete))
        .route("/{id}", get(view))
        .route("/{id}/name", post(update_name))
        .route("/{id}/add-station", post(add_station))
        .route(
            "/{route_id}/delete-station/{station_id}",
            get(delete_station),
        )
        .route(
            "/{route_id}/move-station-higher/{station_id}",
            get(move_station_higher),
        )
}
